/*
  Functions and classes which track the conntracks for l7-filter.

  By Ethan Sommer <sommere@users.sf.net> and Matthew Strait
  <quadong@users.sf.net>, (C) 2006-2007
  http://l7-filter.sf.net

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version
  2 of the License, or (at your option) any later version.
  http://www.gnu.org/licenses/gpl.txt

  Based on ctnl_test.c from libnetfilter-conntrack 0.0.31 (C) 2005 by
  Pablo Neira Ayuso <pablo@eurodev.net>
   and
  Based on nfqnl_test.c from libnetfilter-queue 0.0.12
*/

using namespace std;
//#define bzero(b,n) memset(b,0,n)
//#define bcopy(b1,b2,n) memmove(b2,b1,n)
//#define bcmp(b1,b2,n) memcmp(b1,b2,n) 

#include <pthread.h>
// for thread affinity in solaris
//#include <sys/types.h>
//#include <sys/processor.h>
//#include <sys/procset.h>
//#include <sys/pset.h>

#include <fstream>
#include <iostream>

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <map>
//#include <sys/types.h>
//#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>
#include "nids.h"

#include <sched.h> // for processor affinity

#include "l7-connection.h"
#include "l7-classify.h"
#include "util.h"
#include "queue-mt.h"

#define int_ntoa(x) inet_ntoa(*((struct in_addr *)&x))
#define MAX 64
#define MAX_PACKET_NUMBER 8    // Try to match up to 8 packets
#define SLEEP_MICRO_SECONDS 20 // When the queue is full or empty, just sleep a while
#define MAX_THREAD_NUMBER 64    // Maximum # of matching threads
#define QUEUE_SIZE 512 

typedef map<string, l7_connection *> l7_map;
typedef QMT<l7_connection *> l7_queue;

l7_map l7_connections;
l7_queue l7_queues[QUEUE_SIZE];

// Probably shouldn't really be global, but it's SO much easier
int maxpackets = 8; // by default.
int clobbermark = 0;

int buf_length;
int num_thread;
int rr_pointer = 0;

//Added by Guangden. they are for buffer charecteristic 
int max_connect1;
int max_connect2;
int connection;
int connection1;
int buffer_size;
int max_buffer;
FILE * output;
FILE * output1;
FILE * output2;

pthread_t threads[MAX];

struct pthread_data
{
    int thread_num;
    l7_queue * p_l7_queue;
    long q_length; // for load balancE
};

pthread_data thread_data[MAX];

static l7_classify * l7_classifier;

// Configurable parameters
extern int verbosity;

extern string l7dir;

extern int clobbermark;

// For thread timing information
struct timeval thread_start_time[MAX_THREAD_NUMBER];
struct timeval thread_end_time[MAX_THREAD_NUMBER];
struct timeval thread_diff_time[MAX_THREAD_NUMBER];
struct timezone thread_tz;

static void handle_sigint(int s)
{
    // was there a reason for this?  It just makes it crash on Ctrl-C for me.
    // delete l7_connection_tracker;
    l7printf(0, "Terminating process.\n");
    for (int i = 0; i < num_thread; i++)
    {
        pthread_kill(threads[i], SIGKILL);
    }
    
    l7printf(0, "L7-filter process terminated.\n");

    exit(0);
}

static void handle_cmdline(string & conffilename, string & tracefilename, int argc, char * *argv)
{
    int dumb = 0;                                       // whether to allow dumb things
    buf_length = MAX_PACKET_NUMBER * MAX_PACKET_LENGTH; // default (8 large packets worth)
    tracefilename = "";
    conffilename = "";
    const char *opts = "t:f:v:h?sb:dn:e:p:m:";

    int done = 0;

    while (!done)
    {
        char c;

        switch (c = getopt(argc, argv, opts))
        {
        case - 1:
            done = 1;
            break;

        case 't':
            tracefilename = optarg;
            break;

        case 'f':
            conffilename = optarg;
            break;

        case 'p':
            l7dir = optarg;
            break;

        case 'e':
            num_thread = atoi(optarg);
            break;

        case 'n':
            maxpackets = strtoll(optarg, 0, 10);

            // never allow maxpackets to be less than one.
            // Allow it to be outside of the range 3-16 only if -d is given
            if (maxpackets == LONG_MIN || maxpackets == LONG_MAX || maxpackets < 1
                || ((maxpackets < 3 || maxpackets > 16) && !dumb))
            {
                cerr << "The number of packets is out of range or you gave me a\n"
                        "non-number. Valid number are between 3 and 16. Or if you\n"
                        "give the -d option before this one, I'll allow any\n"
                        "positive integer.\n";

                exit(1);
            }
            
            buf_length = maxpackets * MAX_PACKET_LENGTH;

            break;

        case 'c':
            clobbermark = 1;
            break;

        case 'v':
            verbosity = atoi(optarg);
            break;

        case 'd':
            dumb = 1;
            cout << "Allowing configurations that are probably ill-advised...\n";
            break;

        case 's':
            verbosity = -1;
            break;

        case 'h':
        case '?':
        default:
            cerr << "l7-filter v" << L7VERSION << ", (C) 2006-2007 Ethan Sommer, Matthew Strait\n"
            "l7-filter comes with ABSOLUTELY NO WARRANTY. This is free software\n"
            "and you may redistribute it under the terms of the GPLv2.\n"
            "\n"
            "Syntax: l7-filter -f configuration_file [options]\n"
            "\n"
            "Options are:\n"
            "-v level\tVerbosity level\n"
            "-e number\tThe number of matching threads.\n"
            "-s\t\tBe silent except in the case of warnings and errors\n"
            "-n packets\tExamine up to this many packets per connection\n"
            "-p path\t\tLook for patterns in path instead of /etc/l7-protocols\n"
            "-d\t\tAllow configurations that are probably ill-advised\n"
	    "-t tracefile\t\tThe input tracefile name\n";

            exit(1);
            break;
        }
    }

    if (conffilename == "")
    {
        cerr << "You must specify a configuration file.  Try 'l7-filter -h'\n";
        exit(1);
    }
}

int gmt2local(time_t t)
{
	register int dt, dir;
	register struct tm *gmt, *loc;
	struct tm sgmt;

	if (t == 0)
		t = time(NULL);
	gmt = &sgmt;
	*gmt = *gmtime(&t);
	loc = localtime(&t);
	dt = (loc->tm_hour - gmt->tm_hour) * 60 * 60 +
	     (loc->tm_min - gmt->tm_min) * 60;

	/*
	 * If the year or julian day is different, we span 00:00 GMT
	 * and must add or subtract a day. Check the year first to
	 * avoid problems when the julian day wraps.
	 */
	dir = loc->tm_year - gmt->tm_year;
	if (dir == 0)
		dir = loc->tm_yday - gmt->tm_yday;
	dt += dir * 24 * 60 * 60;

	return (dt);
}

string get_time_str(struct timeval *tvp)
{
    char buf[256];
    int size;
    int s;
    struct tm *tm;
    time_t time;
    int thiszone;
    thiszone = gmt2local(0);

    s = (tvp->tv_sec + thiszone) % 86400;
    time = (tvp->tv_sec + thiszone) - s;
    tm = gmtime (&time);
    
    if (tm == NULL)
    {
        return "Fail to get time.";
    }
    else
    {
        size = sprintf(buf, "%04d-%02d-%02d ",
                 tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday);
    }
    sprintf(buf + size, "%02d:%02d:%02d.%06u",
            s / 3600, (s % 3600) / 60, s % 60, (unsigned)tvp->tv_usec);
    
    string str = buf;
    return str;
}

void schedule(l7_connection * p_l7_conn)
{
    // Here it is round robin like policy to distribute the packets to all available regular expression engines
 
    while (1) {
	if (rr_pointer >= num_thread)
	    rr_pointer=0;
//	cout<<"thread: "<<p_l7_conn->get_thread_num()<<endl;
//	cout<<"pck #: "<<thread_data[rr_pointer].p_l7_queue->GetEntryNum()<<endl;
	if (thread_data[rr_pointer].p_l7_queue->Push(p_l7_conn) == true)
        {
            p_l7_conn->set_thread_num(rr_pointer++);
            break;
        }
    }

    // Here is an connection-based scheduler to more efficiently utilize the multi-core
    // new connection
/*    if (p_l7_conn->get_thread_num() == 0)
    {
        while (1)
        {
	    // This is a load balance to find the most idling thread for this new connection
	    int least=1;
            for (int i=0; i<num_thread; i++)
	    {
		if (thread_data[least].p_l7_queue->GetEntryNum() > thread_data[i].p_l7_queue->GetEntryNum())
		    least = i;
	    }
	    if (thread_data[least].p_l7_queue->Push(p_l7_conn) == true)
	    {
                p_l7_conn->set_thread_num(least);
                break;
            }
        }
    }
    else
    {
        while (thread_data[p_l7_conn->get_thread_num()].p_l7_queue->Push(p_l7_conn) == false)
        {
            usleep(SLEEP_MICRO_SECONDS);
        }

	// load balance to the thread with the least workload if the desired thread is saturated
	if (thread_data[p_l7_conn->get_thread_num()].p_l7_queue->GetEntryNum() == MAX_PACKET_NUMBER)
	{
	    int least=0;
            for (int i=0; i<num_thread; i++)
            {
                if (thread_data[least].p_l7_queue->GetEntryNum() > thread_data[i].p_l7_queue->GetEntryNum())
                    least = i;
            }
            if (thread_data[least].p_l7_queue->Push(p_l7_conn) == true)
            {
                p_l7_conn->set_thread_num(least);
            }
	}
	else 
	    thread_data[p_l7_conn->get_thread_num()].p_l7_queue->Push(p_l7_conn);
    }
*/
}
/*
signed long core_weights[8] = {0, 0, 0, 0, 0, 0, 0, 0};
signed long pipe_weights[8][2]= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
signed long thread_weights[8][2][4] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; 
int core_len[8] = {0, 0, 0, 0, 0, 0, 0, 0};
int pipe_len[8][2]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int thread_len[8][2][4] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int core_weight_max_id = 0;
int core_len_min_id =0; 
#define CORE_LEN_THRESHOLD 100 // usually packets are 200 bytes on average, 
			       // and to classify a conn need less than 2 packets 
int core_select =0;
int pipe_weight_max_id = 0;
int pipe_len_min_id =0;
#define PIPE_LEN_THRESHOLD 100 // usually packets are 200 bytes on average,
                               // and to classify a conn need less than 2 packets
int pipe_select = 0;
int thread_weight_max_id = 0;
int thread_len_min_id =0;
#define THREAD_LEN_THRESHOLD 100 // usually packets are 200 bytes on average,
                               // and to classify a conn need less than 2 packets
int thread_select = 0;
#include <math.h>
*/
/*******************************************************************
        To computer the theory adjust value based on Ross's paper.
int HRW(int conn_id, unsigned int buff_len)
{
        // Schedule to a Core first
        for (int i=0; i<8; i++)
        {
                // find the minimum core_len
                if (core_len[core_len_min_id] > core_len[i])
                        core_len_min_id = i;
        }
        // recursive calculation
        double p[8];// sum(p_i)=1; it's a influence ratio
        double sum=0.0;
        for (int i=0; i<8; ++i) {
                sum+=(core_len[core_len_min_id]/core_len[i]);
        }
        for (int i=0; i<8; ++i) {
                p[i] = (core_len[core_len_min_id]/core_len[i])/sum;
        }
        double x[8]; // x[i] is the theoretically correct adjustement for PU[i]
        x[0]=pow(8*p[0], 1/8);
        for (int i=1; i<8; ++i) {
                double mul_xi=1.0;
                for (int j=0; j<i; ++j)
                        mul_xi*=x[j];
                x[i]=pow(pow(x[i-1], 8-i)+(N-i)*(p[i]-p[i-1])/mul_xi, 1/(8-i));
        }
	// apply x[0..7] to all the cores
        for (int i=0; i<8; i++)
        {
                core_weights[i] = x[i]* abs((signed int)(1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (core_weights[core_weight_max_id] < core_weights[i])
                        core_weight_max_id = i;
        }
        core_select = core_weight_max_id;
        core_len[core_select]+=(int)buff_len;

        // Schedule to a Pipe on core_select
        for (int i=0; i<2; i++)
        {
                // find the minimum pipe_len
                if (pipe_len[core_select][pipe_len_min_id] > pipe_len[core_select][i])
                        pipe_len_min_id = i;
        }
        // recursive calculation
        sum=0.0; // reset sum
        for (int i=0; i<2; ++i) {
                sum+=(pipe_len[pipe_len_min_id]/pipe_len[i]);
        }
        for (int i=0; i<2; ++i) {
                p[i] = (pipe_len[pipe_len_min_id]/pipe_len[i])/sum;
        }
        x[0]=pow(2*p[0], 1/2);
        for (int i=1; i<2; ++i) {
                double mul_xi=1.0;
                for (int j=0; j<i; ++j)
                        mul_xi*=x[j];
                x[i]=pow(pow(x[i-1], 2-i)+(N-i)*(p[i]-p[i-1])/mul_xi, 1/(2-i));
        }
        // apply x[0..1] to all the cores

        for (int i=0; i<2; i++)
        {
                pipe_weights[core_select][i] = x[i] * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (pipe_weights[core_select][pipe_weight_max_id] < pipe_weights[core_select][i])
                        pipe_weight_max_id = i;
        }
        pipe_select = pipe_weight_max_id;
        pipe_len[core_select][pipe_select]+= (int)buff_len;

        // Schedule to a thread on pipe_select on core_select
        for (int i=0; i<4; i++)
        {
                // find the minimum thread_len
                if (thread_len[core_select][pipe_select][thread_len_min_id] > thread_len[core_select][pipe_select][i])
                        thread_len_min_id = i;
        }
        // recursive calculation
        sum=0.0; // reset sum
        for (int i=0; i<4; ++i) {
                sum+=(thread_len[thread_len_min_id]/thread_len[i]);
        }
        for (int i=0; i<4; ++i) {
                p[i] = (thread_len[thread_len_min_id]/thread_len[i])/sum;
        }
        x[0]=pow(4*p[0], 1/4);
        for (int i=1; i<4; ++i) {
                double mul_xi=1.0;
                for (int j=0; j<i; ++j)
                        mul_xi*=x[j];
                x[i]=pow(pow(x[i-1], 4-i)+(N-i)*(p[i]-p[i-1])/mul_xi, 1/(4-i));
        }
        // apply x[0..3] to all the cores

        for (int i=0; i<4; i++)
        {
                thread_weights[core_select][pipe_select][i] = x[i] * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (thread_weights[core_select][pipe_select][thread_weight_max_id] < thread_weights[core_select][pipe_select][i])
                        thread_weight_max_id = i;
        }
        thread_select = thread_weight_max_id;
        thread_len[core_select][pipe_select][thread_select]+= (int)buff_len;
//cout<<"\tthread select ="<<thread_select<<endl;
        return core_select*8 + pipe_select*4 + thread_select;

}
********************************************/


// Now it's the Adjusted Hightest Random Weight (AHRW) based scheduling
// This version of HRW actually adjusts the HRW weight on all the PUs
/*int HRW(int conn_id, unsigned int buff_len)
{	
	// Schedule to a Core first
	for (int i=0; i<8; i++)
	{
		// find the minimum core_len
		if (core_len[core_len_min_id] > core_len[i])
			core_len_min_id = i;
	}
	for (int i=0; i<8; i++)		
	{
		double adjust = 0.0;
		if (core_len[core_len_min_id] != 0 && core_len[i]!=0)
			adjust = core_len[core_len_min_id] / core_len[i];
		else
			adjust = 1;	
		core_weights[i] = adjust * abs((signed int)(1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
		if (core_weights[core_weight_max_id] < core_weights[i])
			core_weight_max_id = i;
	}	
	core_select = core_weight_max_id;
	core_len[core_select]+=(int)buff_len;
//cout<<"core select ="<<core_select<<endl;

        // Schedule to a Pipe on core_select
        for (int i=0; i<2; i++)
        {
                // find the minimum pipe_len
                if (pipe_len[core_select][pipe_len_min_id] > pipe_len[core_select][i])
                        pipe_len_min_id = i;
        }
        for (int i=0; i<2; i++)
        {
                double adjust = 0.0;
                if (pipe_len[core_select][pipe_len_min_id] != 0 && pipe_len[core_select][i] != 0)
                        adjust = pipe_len[core_select][pipe_len_min_id] / pipe_len[core_select][i];
                else
                        adjust = 1;
                pipe_weights[core_select][i] = adjust * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (pipe_weights[core_select][pipe_weight_max_id] < pipe_weights[core_select][i])
                        pipe_weight_max_id = i;
        }
	pipe_select = pipe_weight_max_id;
	pipe_len[core_select][pipe_select]+= (int)buff_len;
//cout<<"\tpipe select ="<<pipe_select<<endl;

        // Schedule to a thread on pipe_select on core_select 
        for (int i=0; i<4; i++)
        {
                // find the minimum thread_len
                if (thread_len[core_select][pipe_select][thread_len_min_id] > thread_len[core_select][pipe_select][i])
                        thread_len_min_id = i;
        }
        for (int i=0; i<4; i++)
        {
                double adjust = 0.0;
                if (thread_len[core_select][pipe_select][thread_len_min_id] != 0 && thread_len[core_select][pipe_select][i] != 0)
                        adjust = thread_len[core_select][pipe_select][thread_len_min_id] / thread_len[core_select][pipe_select][i];
                else
                        adjust = 1;
                thread_weights[core_select][pipe_select][i] = adjust * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (thread_weights[core_select][pipe_select][thread_weight_max_id] < thread_weights[core_select][pipe_select][i])
                        thread_weight_max_id = i;
        }
        thread_select = thread_weight_max_id;
	thread_len[core_select][pipe_select][thread_select]+= (int)buff_len;
//cout<<"\tthread select ="<<thread_select<<endl;
	return core_select*8 + pipe_select*4 + thread_select;

}
*/

/*
// Now it's a Hightest Random Weight (HRW) based scheduling that only involves the min_len
// PU and the HRW selected PU. A better version is presented in the function above.
int HRW(int conn_id, unsigned int buff_len)
{
cout << "test" << endl;
        // Schedule to a Core first
        for (int i=0; i<8; i++)
        {
                core_weights[i] = abs((signed int)(1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (core_weights[core_weight_max_id] < core_weights[i])
                        core_weight_max_id = i;
                if (core_len[core_len_min_id] > core_len[i])
                        core_len_min_id = i;
        }
        if (core_len[core_weight_max_id] - core_len[core_len_min_id] >= 0)
                core_select = core_len_min_id;
        else
                core_select = core_weight_max_id;
        core_len[core_select]+=(int)buff_len;
//cout<<"core select ="<<core_select<<endl;
        // Schedule to a Pipe on core_select
        for (int i=0; i<2; i++)
        {
                pipe_weights[core_select][i] = abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (pipe_weights[core_select][pipe_weight_max_id] < pipe_weights[core_select][i])
                        pipe_weight_max_id = i;
                if (pipe_len[core_select][pipe_len_min_id] > pipe_len[core_select][i])
                        pipe_len_min_id = i;
        }
        if (pipe_len[core_select][pipe_weight_max_id] -
                pipe_len[core_select][pipe_len_min_id] >= 0)
                pipe_select = pipe_len_min_id;
        else
                pipe_select = pipe_weight_max_id;
        pipe_len[core_select][pipe_select]+= (int)buff_len;
//cout<<"\tpipe select ="<<pipe_select<<endl;
        // Schedule to a thread on pipe_select on core_select
        for (int i=0; i<4; i++)
        {
                thread_weights[core_select][pipe_select][i] = abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (thread_weights[core_select][pipe_select][thread_weight_max_id] < thread_weights[core_select][pipe_select][i])
                        thread_weight_max_id = i;
                if (thread_len[core_select][pipe_select][thread_len_min_id] > thread_len[core_select][pipe_select][i])
                        thread_len_min_id = i;
        }
        if (thread_len[core_select][pipe_select][thread_weight_max_id] -
                thread_len[core_select][pipe_select][thread_len_min_id] >= 0)
                thread_select = thread_len_min_id;
        else
                thread_select = thread_weight_max_id;
        thread_len[core_select][pipe_select][thread_select]+= (int)buff_len;
//cout<<"\tthread select ="<<thread_select<<endl;
        return core_select*8 + pipe_select*4 + thread_select;

}
*/

pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
// the HRW scheduler
/*void HRW_schedule(l7_connection * p_l7_conn)
{
	pthread_mutex_lock( &mutex1 );
	int thread_id = HRW(p_l7_conn->ID, p_l7_conn->get_buffer_cur_length());
	thread_data[thread_id].p_l7_queue->Push(p_l7_conn); 
	pthread_mutex_unlock( &mutex1 );
}
*/

int conn_ID = 0;
void tcp_callback(struct tcp_stream *a_tcp, void * *this_time_not_needed)
{
    char direction[16];
    l7_connection * p_l7_conn = NULL;

    if (a_tcp->nids_state == NIDS_JUST_EST)
    {
        // connection described by a_tcp is established
        // here we decide, if we wish to follow this stream
        // sample condition: if (a_tcp->addr.dest!=23) return;
        // in this simple app we follow each stream, so..
        a_tcp->client.collect++;     // we want data received by a client
        a_tcp->server.collect++;     // and by a server, too
        a_tcp->server.collect_urg++; // we want urgent data received by a server
        a_tcp->client.collect_urg++; // if we don't increase this value,
                                     // we won't be notified of urgent data
                                     // arrival

        p_l7_conn = new l7_connection(maxpackets);

	connection=connection+1;
	//fprintf(output, "%d\n", connection);
	

	connection1=connection1+1;
	//fprintf(output1, "%d\n", connection1);
	
	if (connection > max_connect1)
		max_connect1 = connection; 
	
	if (connection1 > max_connect2)
//		max_connect1 = connection; 
		max_connect2 = connection1;
        
	if (p_l7_conn == NULL)
        {
            fprintf(stderr, "Cannot new l7_connection instance, out of memory!\n");
            exit(1);
        }

        string key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.saddr), int_ntoa(a_tcp->addr.daddr),
                                       a_tcp->addr.source, a_tcp->addr.dest);
        p_l7_conn->set_key(key);
	p_l7_conn->ID = conn_ID++; // set the connection ID
        l7_connections[key] = p_l7_conn;

        l7printf(2, "%s (established)\n", key.c_str());
        return;
    }

    if (a_tcp->nids_state == NIDS_CLOSE)
    {
        // connection has been closed normally
        string key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.saddr), int_ntoa(a_tcp->addr.daddr),
                                       a_tcp->addr.source, a_tcp->addr.dest);
        p_l7_conn = l7_connections[key];

        if (p_l7_conn == NULL)
        {
            key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.daddr), int_ntoa(a_tcp->addr.saddr), a_tcp->addr.dest,
                                    a_tcp->addr.source);
            p_l7_conn = l7_connections[key];
        }

        if (p_l7_conn != NULL)
        {
            // delete p_l7_conn;
            l7_connections.erase(l7_connections.find(key));
	    
	    connection--;
	    //fprintf(output, "%d\n", connection);

	    if (!p_l7_conn->flag)
	    {
		connection1--;
	    //	fprintf(output1, "%d\n", connection1);
		p_l7_conn->flag =1 ;
	    }
        }
	else key="No connection found!"; // connection does not exit.
        l7printf(2, "%s (closing)\n", key.c_str());
        return;
    }

    if (a_tcp->nids_state == NIDS_RESET)
    {
        // connection has been closed by RST
        string key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.saddr), int_ntoa(a_tcp->addr.daddr),
                                       a_tcp->addr.source, a_tcp->addr.dest);
        p_l7_conn = l7_connections[key];

        if (p_l7_conn == NULL)
        {
            key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.daddr), int_ntoa(a_tcp->addr.saddr), a_tcp->addr.dest,
                                    a_tcp->addr.source);
            p_l7_conn = l7_connections[key];
        }

        if (p_l7_conn != NULL)
        {
            // delete p_l7_conn;
	    connection--;
	    //fprintf(output, "%d\n", connection);
           
 	    if (!p_l7_conn->flag)
            {
		connection1--;
                //fprintf(output1, "%d\n", connection1);
                p_l7_conn->flag =1 ;
            }

            l7_connections.erase(l7_connections.find(key));
        }
	else key="No connection found!"; // connection does not exit.
        l7printf(2, "%s (reset)\n", key.c_str());
        return;
    }
    
    if (a_tcp->nids_state == NIDS_TIMED_OUT)
    {
        // connection has been timed out
        string key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.saddr), int_ntoa(a_tcp->addr.daddr),
                                       a_tcp->addr.source, a_tcp->addr.dest);
        p_l7_conn = l7_connections[key];

        if (p_l7_conn == NULL)
        {
            key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.daddr), int_ntoa(a_tcp->addr.saddr), a_tcp->addr.dest,
                                    a_tcp->addr.source);
            p_l7_conn = l7_connections[key];
        }

        if (p_l7_conn != NULL)
        {
            // delete p_l7_conn;
	    connection--;
	    //fprintf(output, "%d\n", connection);

	    if (!p_l7_conn->flag)
            {
		connection1--;
              //  fprintf(output1, "%d\n", connection1);
                p_l7_conn->flag =1 ;
            }
            l7_connections.erase(l7_connections.find(key));
        }
	else key="No connection found!"; // connection does not exit.
        l7printf(2, "%s (timeout)\n", key.c_str());
        return;
    }

    if (a_tcp->nids_state == NIDS_DATA)
    {
        // new data has arrived; gotta determine in what direction
        // and if it's urgent or not

        struct half_stream * hlf = NULL;
        unsigned int mark = UNTOUCHED;

        string key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.saddr), int_ntoa(a_tcp->addr.daddr),
                                       a_tcp->addr.source, a_tcp->addr.dest);
        p_l7_conn = l7_connections[key];

        if (p_l7_conn == NULL)
        {
            key = get_conntrack_key("TCP", int_ntoa(a_tcp->addr.daddr), int_ntoa(a_tcp->addr.saddr), a_tcp->addr.dest,
                                    a_tcp->addr.source);
            p_l7_conn = l7_connections[key];
        }

        if (p_l7_conn == NULL)
        {
            l7printf(2, "Got packet, had no connection:\t%s\n", key.c_str());
        }

        if (a_tcp->server.count_new_urg)
        {
            // new byte of urgent data has arrived
            return;
        }

        // We don't have to check if urgent data to client has arrived,
        // because we haven't increased a_tcp->client.collect_urg variable.
        // So, we have some normal data to take care of.
        if (a_tcp->client.count_new > 0)
        {
            // new data for client
            hlf = &a_tcp->client; // from now on, we will deal with hlf var,
            sprintf(direction, " (<-)");
        }
        else if (a_tcp->server.count_new > 0)
        {
            hlf = &a_tcp->server; // analogical
            sprintf(direction, " (->)");
        }
        else
        {
            return;
        }

        // write(2, hlf->data, hlf->count_new);
        // fprintf(stderr, "\n");

        if (p_l7_conn != NULL)
        {
            l7printf(2, "%s%s\n", key.c_str(), direction);

            mark = p_l7_conn->get_mark();

            if (mark == UNTOUCHED || mark == NO_MATCH_YET)
            {
                if (p_l7_conn->get_packet_number() <= MAX_PACKET_NUMBER &&
                    p_l7_conn->set_new_packet(hlf->data, hlf->count_new) == true)
                {
                    if (verbosity >= 2)
                    {
                        write(1, p_l7_conn->get_buffer(), p_l7_conn->get_buffer_cur_length());
                        //printf("\n");
			l7printf(2, "\n");
                    }
                    schedule(p_l7_conn);
// Using the new HRW scheduler
//			HRW_schedule(p_l7_conn);			
                }
		else {
	            if (!p_l7_conn->flag)
        	    {	
		        connection1--;
                        //fprintf(output1, "%d\n", connection1);
                        p_l7_conn->flag =1 ;
//		        fprintf(output2, "%d\n", p_l7_conn->get_buffer_cur_length());
            	    }
		}
            }
            else {
	        if (!p_l7_conn->flag)
             	{
		    connection1--;
                    //fprintf(output1, "%d\n", connection1);
                    p_l7_conn->flag =1 ;
//		    fprintf(output2, "%d\n", p_l7_conn->get_buffer_cur_length());
            	}
		l7printf(2, "This connection has been marked as %s.\n",
                        	 l7_classifier->get_pattern_name(mark).c_str());
            }
        }
        else
        {
            l7printf(2, "l7_connection instance is not valid.\n");
        }
    }

    return;
}

void udp_callback(struct tuple4 * addr, char * buf, int len, struct ip * iph)
{
    char direction[16];
    l7_connection * p_l7_conn = NULL;
    
    string key = get_conntrack_key("UDP", int_ntoa(addr->saddr), int_ntoa(addr->daddr),
                                   addr->source, addr->dest);
    p_l7_conn = l7_connections[key];
    sprintf(direction, " (->)");

    if (p_l7_conn == NULL)
    {
        key = get_conntrack_key("UDP", int_ntoa(addr->daddr), int_ntoa(addr->saddr), addr->dest,
                                addr->source);
        p_l7_conn = l7_connections[key];
    }

    // New connection found
    if (p_l7_conn == NULL)
    {
        p_l7_conn = new l7_connection(maxpackets);

	connection++;
	if (connection > max_connect1)
		max_connect1 = connection;
    
	connection1++;
	if (connection1 > max_connect2)
		max_connect2 = connection1;

	//fprintf(output, "%d\n", connection);
	//fprintf(output1, "%d\n", connection1);


        if (p_l7_conn == NULL)
        {
            fprintf(stderr, "Cannot new l7_connection instance, out of memory!\n");
            exit(1);
        }
        
        key = get_conntrack_key("UDP", int_ntoa(addr->saddr), int_ntoa(addr->daddr),
                                addr->source, addr->dest);
        p_l7_conn->set_key(key);
	p_l7_conn->ID = conn_ID++; // set the connection ID
        l7_connections[key] = p_l7_conn;
        sprintf(direction, " (->)");

        l7printf(2, "%s (established)\n", key.c_str());
    }
    else // Original connection found
    {
    }

    if (len <= 0)
    {
        return;
    }

    l7printf(2, "%s%s\n", key.c_str(), direction);

    unsigned int mark = p_l7_conn->get_mark();
    if (mark == UNTOUCHED || mark == NO_MATCH_YET)
    {        
        if (p_l7_conn->get_packet_number() <= MAX_PACKET_NUMBER &&
            p_l7_conn->set_new_packet(buf, len) == true)
        {
            if (verbosity >= 2)
            {
                write(1, p_l7_conn->get_buffer(), p_l7_conn->get_buffer_cur_length());
                printf("\n");
            }

            schedule(p_l7_conn);
// Using the new HRW scheduler
//                        HRW_schedule(p_l7_conn);

        }
	else {
	    if (!p_l7_conn->flag) {
	        connection1--;
                //fprintf(output1, "%d\n", connection1);
		p_l7_conn->flag = 1;
//		fprintf(output2, "%d\n", p_l7_conn->get_buffer_cur_length());
	    }
        }
    }
    else
    {
	if (!p_l7_conn->flag)    
 	 {
	     connection1--;
	     //fprintf(output1, "%d\n", connection1);
	     p_l7_conn->flag = 1;
//	     fprintf(output2, "%d\n", p_l7_conn->get_buffer_cur_length());
	 }
	 l7printf(2, "This connection has been marked as %s.\n",
                 l7_classifier->get_pattern_name(mark).c_str());
    }
}

int n_connection=0;
// Mark the overall queue length per thread
long ql[MAX_THREAD_NUMBER];

static void * regex_matching_thread(void * data)
{
    pthread_data * pdata = (struct pthread_data *)(data);

    /*Now start to the regular experession work*/
    l7printf(0, "Thread %d starts to work right now.\n", pdata->thread_num);

// If we comment out the line below, we are using the default scheduler in Solaris
// which load balance across all the cores first, then balance between the 2 
// pipelines inside a core.
//    processor_bind(P_LWPID, P_MYID, pdata->thread_num, NULL);

    while (1)
    {
        l7_connection * p_l7_connection = NULL;
        
        while(pdata->p_l7_queue->Pop(p_l7_connection) == false)
        {
            usleep(SLEEP_MICRO_SECONDS);
        }

        if (p_l7_connection == NULL)
        {
            break;
        }
        
        unsigned int mark = p_l7_connection->get_mark();
        if (mark != UNTOUCHED && mark != NO_MATCH_YET && mark != NO_MATCH)
        {
            continue;
        }
//cout<<"buffer size:"<<p_l7_connection->get_buffer_cur_length()<<endl;
        p_l7_connection->reassembly_new_packet();

// accumulate the thread queue length
int cur_len = p_l7_connection->get_buffer_cur_length();
pdata->q_length += cur_len;

        char *buffer = p_l7_connection->get_buffer();
        mark = l7_classifier->classify(buffer);
// add some additional workload (TEST)
float a=1.0;
float b=2.3;
for (int i=0; i<170000; i++)
        a = a*(b++);

        p_l7_connection->set_mark(mark);
	n_connection++;
        
        if (mark != UNTOUCHED && mark != NO_MATCH_YET && mark != NO_MATCH)
        {
    		if (p_l7_connection->get_packet_number() < MAX_PACKET_NUMBER )
			fprintf(output2, "%d\n", p_l7_connection->get_buffer_cur_length());
            	l7printf(1, "Match %s:\t%s by thread %d.\n",
                     l7_classifier->get_pattern_name(mark).c_str(),
                     p_l7_connection->get_key().c_str(),
                     pdata->thread_num);
        
       }
	if (p_l7_connection->get_packet_number() == MAX_PACKET_NUMBER) {
		fprintf(output2, "%d\n", p_l7_connection->get_buffer_cur_length());
	}
// reduce the queue size after the connection buffer is processed.
// for HRW
/*	int c_id = pdata->thread_num/8;
	int p_id = (pdata->thread_num%8)/4;
	int t_id = (pdata->thread_num%8)%4;
	thread_len[c_id][p_id][t_id] -= cur_len;
	pipe_len[c_id][p_id] -= cur_len;
	core_len[c_id] -= cur_len;
*/	
    }

    l7printf(0, "Thread %d finished.\n", pdata->thread_num);

    // Get thread ending information
    bzero((char *) &thread_tz, sizeof(thread_tz));
    gettimeofday(&thread_end_time[pdata->thread_num], &thread_tz);
    TIMERSUB(&thread_end_time[pdata->thread_num], 
		&thread_start_time[pdata->thread_num], 
		&thread_diff_time[pdata->thread_num]);
    l7printf(0, "Running time for thread %d was %lu.%lu seconds\n",
        	pdata->thread_num, 
		(unsigned long)thread_diff_time[pdata->thread_num].tv_sec, 
		(unsigned long)thread_diff_time[pdata->thread_num].tv_usec);

    // calculate the overall thread queue length
    ql[pdata->thread_num] = pdata->q_length;

    pthread_exit(NULL);
}

static int init_threads()
{
    int i = 0;
    int res = 0;

    for (i = 0; i < num_thread; i++)
    {
        if (l7_queues[i].Init(1<<16) == false)
        {
            return 1;
        }

        thread_data[i].thread_num = i;
        thread_data[i].p_l7_queue = &l7_queues[i];

	// Get timing information for each thread.
	bzero((char *) &thread_tz, sizeof(thread_tz));
	gettimeofday(&thread_start_time[i], &thread_tz);

        res = pthread_create(&threads[i], NULL, regex_matching_thread, (void *) & thread_data[i]);
        
        if (res != 0)
        {
            return res;
        }
    }

    return res;
}

// used for papi to count L1 instruction cache miss rate
// #include <papi.h>
//#define NUM_EVENTS 2
int main(int argc, char * *argv)
{
    // Start counting
//    int Events[NUM_EVENTS] = {/*PAPI_L1_TCM, PAPI_L1_TCA,*/ PAPI_L2_TCM, PAPI_L2_TCA/*, PAPI_L1_ICH, PAPI_L1_ICR, PAPI_L1_ICW*/};
//    long_long values[NUM_EVENTS];
    /* Start counting events */
/*    if (PAPI_start_counters(Events, NUM_EVENTS) != PAPI_OK)
    {
        printf("error at starting counter.\n");
        exit (0);
    }
*/

    int res;
    string tracefilename;
    string conffilename;
    
    struct timeval start_time;
    struct timeval end_time;
    struct timeval diff_time;
    struct timezone tz;

    handle_cmdline(conffilename, tracefilename, argc, argv);

    signal(SIGINT, handle_sigint);

    l7_classifier = new l7_classify(conffilename);

    /* Create the threads to do regular expression there. */

    res = init_threads();

    if (res)
    {
        l7printf(0, "Failed to creat the regular expression threads\n");
        exit(1);
    }

    if (tracefilename == "")
    {
        nids_params.filename = NULL;
    }
    else
    {
        nids_params.filename = (char *)tracefilename.c_str();
    }
    
    nids_params.tcp_workarounds = true;

    if (!nids_init())
    {
        l7printf(0, "%s\n", nids_errbuf);
        exit(1);
    }

// Bind the current thread (preprocessing thread) to vcpu#0
// From our experiment results, we show that one vcpu is not enough for preprocessing
//	processor_bind(P_LWPID, P_MYID, 0, NULL);
// Try use a set of processors
/*	psetid_t newpset;	
	pset_create(&newpset);
	pset_bind(newpset, P_PID, P_MYID, NULL);
	pset_assign(newpset, 0, NULL);
	pset_assign(newpset, 8, NULL);
*/
// Add an additional vcpu for preprocessing
//	pset_assign(P_MYID, 8, NULL);
//	pset_bind(P_MYID, P_LWPID, P_MYID, NULL);
//	pset_bind(newpset, P_LWPID, P_MYID, NULL); 

    nids_register_tcp((void *)tcp_callback);
    nids_register_udp((void *)udp_callback);
    
//	pset_destroy(newpset);

    bzero((char *) &tz, sizeof(tz));
    gettimeofday(&start_time, &tz);
    output = fopen("memory", "w"); 
    if (!output)
	exit(0);

    output1 = fopen("memory-buffer", "w"); 
    if (!output1)
	exit(0);

    output2 = fopen("buffer-size", "w"); 
    if (!output2)
	exit(0);

    nids_run();

    // Notify all subthreads to exit
    for (int i = 0; i < num_thread; i++)
    {
        while(thread_data[i].p_l7_queue->Push(NULL) == false)
        {
            usleep(SLEEP_MICRO_SECONDS);
        }
        l7printf(0, "Waiting thread %d to finish.\n", i);
        pthread_join(threads[i], NULL);

    }
   
   
    bzero((char *) &tz, sizeof(tz));
    gettimeofday(&end_time, &tz);
	TIMERSUB(&end_time, &start_time, &diff_time);
	l7printf(0, "Running time was %lu.%lu seconds\n", 
             (unsigned long)diff_time.tv_sec, (unsigned long)diff_time.tv_usec);
    l7_classifier->statistics();
    printf("the maximum connection table is %d\n", max_connect1);    
    printf("the maximum buffer table is %d\n", max_connect2);    
    printf("the number of connections is: %d\n", n_connection);

        // Caculate thread queue length
        FILE * qlen;
        qlen = fopen("qlen", "w");
        if (!qlen)
            exit(0);
        for (int i=0; i<MAX_THREAD_NUMBER; i++)
                fprintf(qlen, "Thread [%d] has a queue lenth of: %ld\n", i, ql[i]);
        fclose(qlen);

    // read counter
/*    if (PAPI_read_counters(values, NUM_EVENTS) != PAPI_OK)
    {
        printf("error at reading counter.\n");
        exit (0);
    }
*/
/*
	PAPI doesn't seem to work with L1 information
    printf("Total number of L1 cache misses is: %lld\n", values[0]);
    printf("Total number of L1 cache acesses is: %lld\n", values[1]);
    printf("L1 Miss rate on the current core is: %f\n", (float)values[0]/values[1]);
*/
/*    printf("Total number of L2 cache misses is: %lld\n", values[0]);
    printf("Total number of L2 cache acesses is: %lld\n", values[1]);
    printf("L2 Miss rate on the current core is: %f\n", (float)values[0]/values[1]);
*/
/*    printf("Total number of L1 instruction cache hits is: %lld\n", values[2]);
    printf("Total number of L1 instruction cache reads is: %lld\n", values[3]);
    printf("Total number of L1 instruction cache writes is: %lld\n", values[4]);
*/
    /* Stop counting events */
/*    if (PAPI_stop_counters(values, NUM_EVENTS) != PAPI_OK)
    {
        printf("error at stopping counter.\n");
        exit (0);
    }
*/
    l7printf(0, "L7-filter process finished.\n");
    fclose(output);
    fclose(output1);
    fclose(output2);
    return 0;
}
